#include <ioavr.h>
#include <intrinsics.h>
#include <string.h> // for memcpy

#define uchar unsigned char


#define DEVICE_VER 6

#define _MCU_CLOCK_FREQUENCY_  18432000
// MCU F in kHz
#define MCUCF 18432

#include "usart.cpp"

struct tReport{
  uchar opcode; // as command + codes for ADC channels
  unsigned int data1;
  unsigned int data2;
};

struct tCommand{
  uchar opcode;
  unsigned int data;
};

// for software calibration and serials
#define EEDataSize 32
__eeprom __no_init uchar EEData[EEDataSize];

__regvar __no_init uchar flags @ 15;

unsigned char KETOvfCnt = 0;

#define FINDATAREADY  1
#define FOUTDATASENT  2
#define FOUTDATAREADY 4
#define FADCRUN       8

#define FlagSet(f) flags = flags | f
#define FlagClear(f) flags = flags & (~f)
#define IsFlag(f) (flags & f)

// port C
#define ADCRbit 0
#define ADCLbit 1

// port D
#define RXbit  0
#define TXbit  1
#define LEDbit 2


static inline void LedToggle(){
  if(PORTD & (1 << LEDbit)) 
    PORTD &= ~(1 << LEDbit); 
  else 
    PORTD |= (1 << LEDbit); 
}


unsigned char CurCh;
void ADC_SelectChannel(uchar ChN){
  ADMUX = (1<<REFS0)|(1<<REFS1)|(0<<ADLAR) | ChN;// ChN pin as input with internal 1.1 V ref
  CurCh = ChN;
}

void ADC_Init(void){ // needed for reference configuration
  //ADMUX = (1<<REFS0)|(1<<REFS1)|(0<<ADLAR)| ADCRbit;// ADC pin as input with internal 1.1 V ref
  ADC_SelectChannel(ADCRbit);
  ADCSRA = (1<<ADEN)|(1<<ADATE)|(1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0); // autotrigger mode with clock / 128
  ADCSRB = 0; // free running mode
  ADCSRA |= (1<<ADIF)|(1<<ADSC); // start ADC
}


void TimersInit(){
  TCCR1A = 0;
  // Normal mode, clk/64 = 288 kHz
  TCCR1B = (1 << CS10) | (1 << CS11);
  // to get microseconds TCNT1 * 4 (for 16 MHz)
  // disable timer interrupts
  TIMSK1 = 0;  
}

/*
#pragma vector = TIMER1_COMPA_vect 
__interrupt void Every10MsInterrupt(void){
  //TimeUs += 10000;
  LedToggle();
}
*/


/* ------------------------------------------------------------------------- */

void WDT_off(void)
{
  __disable_interrupt();
  __watchdog_reset();
  /* Clear WDRF in MCUSR */
  MCUSR &= ~(1<<WDRF);
  /* Write logical one to WDCE and WDE */
  /* Keep old prescaler setting to prevent unintentional time-out */
  WDTCSR |= (1<<WDCE) | (1<<WDE);
  /* Turn off WDT */
  WDTCSR = 0x00;
  //__enable_interrupt();
}

void WDT_on(void)
{
  __disable_interrupt();
  __watchdog_reset();
  /* Set WDRF in MCUSR */
  MCUSR |= (1<<WDRF);
  /* Start timed sequence */
  WDTCSR |= (1<<WDCE) | (1<<WDE);
  /* Set new prescaler(time-out) value = 1024K (1048576) cycles (8.0 s), because wdt has own oscillator 128kHz*/
  WDTCSR = (1<<WDE) | (1<<WDP3) | (1<<WDP0);
  //__enable_interrupt();
}


void PutReport(tReport *rep){
  uchar sz = sizeof(tReport);
  uchar *pb = (uchar*)rep;
  while(sz--){
    putchar(*(pb++));
  }  
}

void RXdata(){       
   tCommand cmd;  
   tReport rep;
   while(rx_counter0 >= sizeof(cmd)){
      cmd.opcode = getchar();
      cmd.data = getchar() | (getchar() << 8);
      
      switch (cmd.opcode){                  
        case 99:        //stop                
          //disable tx of all ADC channels             
          FlagClear(FADCRUN);                                 
        break;
        
        case 101:       //ping            
            rep.opcode = 101;
            rep.data1 = cmd.data;
            PutReport(&rep);  
            //LedToggle(); 
        break;
        
        case 102:       // version                  
            rep.opcode = 102;
            rep.data1 = DEVICE_VER; 
            PutReport(&rep);    
        break;  
        
        case 103:       // keepalive
            // if was not keepalive during some period - reset device state
            // do nothing
        break;
        
        case 104: // eewrite - save 1 byte of user data in eeprom
            {
              uchar addr, data;
              addr = (cmd.data >> 8) & 0xFF;
              data = cmd.data & 0xFF;
              if(addr < EEDataSize){
                EEData[addr] = data;
              }
            }
        break;
        
        case 105: // eeread - read 1 byte of user data from eeprom
            {
              rep.opcode = 105;
              uchar addr, data;
              addr = cmd.data; 
              rep.data1 = addr;
              if(addr < EEDataSize){
                data = EEData[addr];
                rep.data2 = data;
              }else{ // error
                rep.data2 = 0xFFFF; // error value
              }
              PutReport(&rep); 
            }
        break;
       
        case 107:       //startadc                   
            FlagSet(FADCRUN);                      
        break;        
        
        default: // ? error ?
          // cleanup buffers                    
          rx_wr_index0 = rx_rd_index0 = rx_counter0 = 0;                          
      } 
      KETOvfCnt = 0;
   }
}


unsigned int ledi;
unsigned int LastWaitT;

void Capture(){  
  if(tx_counter0 > 0 || rx_counter0 >= sizeof(tCommand)) return;
    
  tReport rep;
  if(CurCh == ADCRbit){ // right is current
    // wait for limit samplerate
    // timer F = 288 kHz. For samplerate 2 kHz wait 144 ticks; 3 kHz -> 96
    unsigned int CurT, dt;
    do{
      CurT = TCNT1;
      if(CurT < LastWaitT) dt = 65535 - LastWaitT + CurT;
      else dt = CurT - LastWaitT;
      if(rx_counter0 > 0) return;
    }while(dt < 72);
    LastWaitT = CurT;
    // read and send sample
    rep.opcode = 5; // Right Channel
    rep.data1 = ADC; 
    rep.data2 = TCNT1;
    ADC_SelectChannel(ADCLbit);// change channel
  }else{ // left
    rep.opcode = 6; // Left Channel
    rep.data1 = ADC; 
    rep.data2 = TCNT1;
    ADC_SelectChannel(ADCRbit);
  }  
  PutReport(&rep);  
  ledi ++;
  if(ledi > 2000){
    LedToggle();      
    ledi = 0;
  }
}



int main(void)
{
    uchar   i;  
   
    DDRC = 0; // all inputs
    PORTC = 0; // no pull-up
    
    DDRB = 0; // all inputs for unused port
    PORTB = 0;
    
    DDRD = (1 << TXbit) | (1 << LEDbit);
    PORTD = 0;
            
    WDT_off();     
    __disable_interrupt();
    
    USARTInit();    
    TimersInit();      
    ADC_Init();    
          
    WDT_on();         
    // start data handling
    __enable_interrupt();   
   
    LedToggle();    
    FlagClear(FADCRUN);
    
    unsigned int CurT;
    unsigned int KeepaliveLastT;
    signed char ledinc = 1;
    unsigned int leddt, ledlastt = 0;
   
    while(1){  /* main event loop */                            
        i++;
        if(i == 0){
          __watchdog_reset();                           
        }  
        
        CurT = TCNT1;
        if(rx_counter0 >= sizeof(tCommand)){
          RXdata();          
        }else{         
          // check keepalive        
          if(KeepaliveLastT > CurT){            
            KETOvfCnt ++;
          }
          KeepaliveLastT = CurT;
                              
          if(KETOvfCnt > 30){ // keepalive timed out            
            KETOvfCnt = 0;        
            FlagClear(FADCRUN);
            rx_wr_index0 = rx_rd_index0 = rx_counter0 = 0;
            tx_wr_index0 = tx_rd_index0 = tx_counter0 = 0;              
          }
          
          if(IsFlag(FADCRUN)){
            Capture();            
            __watchdog_reset();
          }else{ // idle mode                 
            if(CurT > ledlastt) leddt = CurT - ledlastt;
            else leddt = 65535 - ledlastt + CurT;
            if(leddt >= 1024){
              ledlastt = CurT;
              ledi += ledinc;
              if(ledi >= 1024){
                ledinc = -2;                                  
              }else if(ledi <= 2){
                ledinc = 2;
              } 
            }
            if(leddt < ledi) PORTD |= (1 << LEDbit);     
            else PORTD &= ~(1 << LEDbit);             
          }
        }        
    }    
}

/* ------------------------------------------------------------------------- */

