// usart.cpp - for immediate including
// 2011.07.18 (c) Dima

#define RXB8 1
#define TXB8 0
#define UPE 2
#define OVR 3
#define FE 4
#define UDRE 5
#define RXC 7


#define FRAMING_ERROR (1<<FE)
#define PARITY_ERROR (1<<UPE)
#define DATA_OVERRUN (1<<OVR)
#define DATA_REGISTER_EMPTY (1<<UDRE)
#define RX_COMPLETE (1<<RXC)

// global func in main.cpp
void RXdata();


// USART Receiver buffer
#define RX_BUFFER_SIZE0 32

unsigned char rx_buffer0[RX_BUFFER_SIZE0];

#if RX_BUFFER_SIZE0 < 256
  unsigned char rx_wr_index0, rx_rd_index0, rx_counter0;
#else
  unsigned int rx_wr_index0, rx_rd_index0, rx_counter0;
#endif

// USART Receiver interrupt service routine
#pragma vector = USART_RX_vect 
__interrupt void usart_rx_isr(void){
  unsigned char status, data;  
  status = UCSR0A;
  data = UDR0;
  
  if((status & (FRAMING_ERROR | PARITY_ERROR | DATA_OVERRUN)) == 0){
    rx_buffer0[rx_wr_index0] = data;
    if (++rx_wr_index0 == RX_BUFFER_SIZE0) rx_wr_index0 = 0;    
    ++rx_counter0;  
    //RXdata(); - don't call it here in ISR - will be lost data!
  }     
}

char getchar(void)
{
  unsigned char data;
  while (rx_counter0 == 0);
  __disable_interrupt();
  data = rx_buffer0[rx_rd_index0];
  if (++rx_rd_index0 == RX_BUFFER_SIZE0) rx_rd_index0 = 0;  
  --rx_counter0; 
  __enable_interrupt();
  return data;
}


// USART Transmitter buffer
#define TX_BUFFER_SIZE0 32
unsigned char tx_buffer0[TX_BUFFER_SIZE0];

#if TX_BUFFER_SIZE0 < 256
unsigned char tx_wr_index0,tx_rd_index0,tx_counter0;
#else
unsigned int tx_wr_index0,tx_rd_index0,tx_counter0;
#endif

// USART Transmitter interrupt service routine
#pragma vector = USART_UDRE_vect 
__interrupt void usart_tx_isr(void){    
  if (tx_counter0){
    --tx_counter0;
    UDR0 = tx_buffer0[tx_rd_index0];
    if (++tx_rd_index0 == TX_BUFFER_SIZE0) tx_rd_index0=0;
  }
}

void putchar(char c){
  while (tx_counter0 == TX_BUFFER_SIZE0);
  __disable_interrupt();
  if (tx_counter0 || ((UCSR0A & DATA_REGISTER_EMPTY) == 0)){
    tx_buffer0[tx_wr_index0] = c;
    if (++tx_wr_index0 == TX_BUFFER_SIZE0) tx_wr_index0 = 0;
    ++tx_counter0;
  }else
    UDR0 = c;   
  __enable_interrupt();
}

void USARTInit(){
  // USART initialization
  // Communication Parameters: 8 Data, 1 Stop, No Parity
  // USART Receiver: On
  // USART Transmitter: On
  // USART0 Mode: Asynchronous
  // USART Baud Rate: 115200
  UCSR0A=0x00;
  UCSR0B=(1<<RXEN0)|(1<<TXEN0)|(1<<UDRIE0)|(1<<RXCIE0);
  UCSR0C=0x06;
  UBRR0H=0x00;
  //UBRR0L=0x09; //115200
  UBRR0L=0x01; //576 000
  tx_wr_index0 = tx_rd_index0 = tx_counter0 = 0;  
  rx_wr_index0 = rx_rd_index0 = rx_counter0 = 0; 
}